@@ -6,6 +6,7 @@
 /*
  * Arnold Robbins, arnold@skeeve.com, May 1993
  * Public Domain
+ * December 2010, move to ANSI C definition for main().
  */
 
 #if HAVE_CONFIG_H
@@ -20,16 +21,20 @@
 #endif
 
 int
-main(argc, argv)
-int argc;
-char **argv;
+main(int argc, char **argv)
 {
     struct passwd *p;
 
     while ((p = getpwent()) != NULL)
+#ifdef ZOS_USS
+        printf("%s:%ld:%ld:%s:%s\n",
+            p->pw_name, (long) p->pw_uid,
+            (long) p->pw_gid, p->pw_dir, p->pw_shell);
+#else
         printf("%s:%s:%ld:%ld:%s:%s:%s\n",
             p->pw_name, p->pw_passwd, (long) p->pw_uid,
             (long) p->pw_gid, p->pw_gecos, p->pw_dir, p->pw_shell);
+#endif
 
     endpwent();
     return 0;
